"""Top2Vec model for topic modeling."""

from typing import List, Dict, Optional, Union, Any, Tuple
import numpy as np
import pandas as pd
import logging
from pathlib import Path
import pickle
import os

try:
    from top2vec import Top2Vec
    TOP2VEC_AVAILABLE = True
except ImportError:
    TOP2VEC_AVAILABLE = False

from meno.modeling.base import BaseTopicModel
from meno.modeling.embeddings import DocumentEmbedding


class Top2VecModel(BaseTopicModel):
    """Top2Vec model for topic modeling.
    
    Parameters
    ----------
    embedding_model : Optional[DocumentEmbedding], optional
        Document embedding model to use, by default None
        If None, Top2Vec will use internal embeddings
    n_topics : Optional[int], optional
        Number of topics to extract, by default None
        If None, Top2Vec automatically determines the number of topics
    min_topic_size : int, optional
        Minimum size of topics, by default 10
    use_gpu : bool, optional
        Whether to use GPU acceleration if available, by default False
        Setting to False (default) ensures CPU-only operation and avoids CUDA dependencies
    umap_args : Optional[Dict], optional
        Arguments for UMAP, by default None
    hdbscan_args : Optional[Dict], optional
        Arguments for HDBSCAN, by default None
    use_custom_embeddings : bool, optional
        Whether to use custom embeddings from DocumentEmbedding, by default True
        If False, Top2Vec will use its own embeddings
    verbose : bool, optional
        Whether to show verbose output, by default True
    
    Attributes
    ----------
    model : Top2Vec
        Trained Top2Vec model
    topics : Dict[int, str]
        Mapping of topic IDs to topic descriptions
    topic_sizes : Dict[int, int]
        Mapping of topic IDs to topic sizes
    """
    
    def __init__(
        self,
        embedding_model: Optional[DocumentEmbedding] = None,
        n_topics: Optional[int] = None,
        min_topic_size: int = 10,
        use_gpu: bool = False,
        umap_args: Optional[Dict] = None,
        hdbscan_args: Optional[Dict] = None,
        use_custom_embeddings: bool = True,
        verbose: bool = True,
    ):
        """Initialize the Top2Vec model."""
        if not TOP2VEC_AVAILABLE:
            raise ImportError(
                "Top2Vec is required for this model. "
                "Install with 'pip install top2vec>=1.0.27'"
            )
            
        self.n_topics = n_topics
        self.min_topic_size = min_topic_size
        self.use_gpu = use_gpu
        self.umap_args = umap_args or {}
        self.hdbscan_args = hdbscan_args or {
            "min_cluster_size": min_topic_size,
            "min_samples": 5,
            "metric": "euclidean",
            "cluster_selection_method": "eom",
        }
        self.use_custom_embeddings = use_custom_embeddings
        self.verbose = verbose
        
        # Set up embedding model if not provided and using custom embeddings
        if embedding_model is None and use_custom_embeddings:
            self.embedding_model = DocumentEmbedding(use_gpu=False)  # Default to CPU
        else:
            self.embedding_model = embedding_model
            
        # Initialize model to None
        self.model = None
        self.topics = {}
        self.topic_sizes = {}
        self.is_fitted = False
    
    def fit(
        self,
        documents: Union[List[str], pd.Series],
        embeddings: Optional[np.ndarray] = None,
    ) -> "Top2VecModel":
        """Fit the Top2Vec model to a set of documents.
        
        Parameters
        ----------
        documents : Union[List[str], pd.Series]
            List or Series of document texts
        embeddings : Optional[np.ndarray], optional
            Pre-computed document embeddings, by default None
            If None and use_custom_embeddings=True, embeddings will be computed
            
        Returns
        -------
        Top2VecModel
            Fitted model
        """
        # Convert pandas Series to list if needed
        if isinstance(documents, pd.Series):
            documents = documents.tolist()
        
        # Handle custom embeddings
        if self.use_custom_embeddings:
            if embeddings is None and self.embedding_model is not None:
                embeddings = self.embedding_model.embed_documents(documents)
                
            # Initialize and fit Top2Vec with custom embeddings
            self.model = Top2Vec(
                documents=documents, 
                document_vectors=embeddings,
                embedding_model=None,  # Custom embeddings already provided
                min_count=1,  # Not used with custom embeddings
                keep_documents=True,
                workers=os.cpu_count() or 1,
                umap_args=self.umap_args,
                hdbscan_args=self.hdbscan_args,
                verbose=self.verbose,
            )
        else:
            # Let Top2Vec handle embeddings
            use_embedding = "doc2vec" if self.embedding_model is None else "universal-sentence-encoder"
            speed = "learn" if self.embedding_model is None else "deep-learn"
            
            self.model = Top2Vec(
                documents=documents,
                embedding_model=use_embedding,
                speed=speed,
                min_count=1,
                keep_documents=True,
                workers=os.cpu_count() or 1,
                umap_args=self.umap_args,
                hdbscan_args=self.hdbscan_args,
                verbose=self.verbose,
            )
            
        # Reduce topics if specified
        if self.n_topics is not None and self.n_topics < len(self.model.topic_sizes):
            self.model.hierarchical_topic_reduction(self.n_topics)
            
        # Store topic information
        self._extract_topic_info()
        
        self.is_fitted = True
        return self
    
    def _extract_topic_info(self) -> None:
        """Extract topic information from the fitted model."""
        # Get topic words and sizes
        topic_words, _, _ = self.model.get_topics()
        topic_sizes = self.model.topic_sizes
        
        # Create topic descriptions
        self.topics = {}
        self.topic_sizes = {}
        
        for i, words in enumerate(topic_words):
            # Skip outlier topic (-1)
            if i < len(topic_sizes):
                # Create topic description from top words
                top_words = words[:5]
                topic_name = f"Topic {i}: {', '.join(top_words)}"
                self.topics[i] = topic_name
                self.topic_sizes[i] = topic_sizes[i]
    
    def transform(
        self,
        documents: Union[List[str], pd.Series],
        embeddings: Optional[np.ndarray] = None,
        top_n: int = 1,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Assign documents to topics.
        
        Parameters
        ----------
        documents : Union[List[str], pd.Series]
            List or Series of document texts
        embeddings : Optional[np.ndarray], optional
            Pre-computed document embeddings, by default None
        top_n : int, optional
            Number of top topics to return for each document, by default 1
            
        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            Tuple of (topic_nums, topic_scores)
            topic_nums has shape (n_documents, top_n)
            topic_scores has shape (n_documents, top_n)
        """
        if not self.is_fitted:
            raise ValueError("Model must be fitted before transform can be called")
            
        # Convert pandas Series to list if needed
        if isinstance(documents, pd.Series):
            documents = documents.tolist()
            
        # Handle custom embeddings
        if self.use_custom_embeddings and embeddings is None and self.embedding_model is not None:
            embeddings = self.embedding_model.embed_documents(documents)
            
        # Get document vectors if using custom embeddings
        if self.use_custom_embeddings and embeddings is not None:
            topic_nums, topic_scores = self.model.get_documents_topics(
                doc_vectors=embeddings,
                num_topics=top_n
            )
        else:
            # Use Top2Vec's built-in methods
            topic_nums, topic_scores = self.model.get_documents_topics(
                doc_ids=None,  # Will use the documents passed to add_documents
                num_topics=top_n
            )
            
        return topic_nums, topic_scores
    
    def add_documents(
        self,
        documents: Union[List[str], pd.Series],
        embeddings: Optional[np.ndarray] = None,
    ) -> None:
        """Add new documents to the trained model.
        
        Parameters
        ----------
        documents : Union[List[str], pd.Series]
            List or Series of document texts to add
        embeddings : Optional[np.ndarray], optional
            Pre-computed document embeddings, by default None
        """
        if not self.is_fitted:
            raise ValueError("Model must be fitted before adding documents")
            
        # Convert pandas Series to list if needed
        if isinstance(documents, pd.Series):
            documents = documents.tolist()
            
        # Handle custom embeddings
        if self.use_custom_embeddings and embeddings is None and self.embedding_model is not None:
            embeddings = self.embedding_model.embed_documents(documents)
            
        # Add documents to the model
        if self.use_custom_embeddings and embeddings is not None:
            self.model.add_documents(
                documents=documents,
                doc_vectors=embeddings
            )
        else:
            self.model.add_documents(documents=documents)
            
        # Update topic information
        self._extract_topic_info()
    
    def search_topics(
        self,
        query: str,
        n_topics: int = 5,
    ) -> List[Tuple[int, str, float]]:
        """Search for topics similar to a query.
        
        Parameters
        ----------
        query : str
            Query string to search for
        n_topics : int, optional
            Number of topics to return, by default 5
            
        Returns
        -------
        List[Tuple[int, str, float]]
            List of tuples (topic_id, topic_name, score)
        """
        if not self.is_fitted:
            raise ValueError("Model must be fitted before searching topics")
            
        # Use Top2Vec's search_topics method
        topic_nums, topic_scores, _ = self.model.search_topics(
            query=query,
            num_topics=n_topics
        )
        
        results = []
        for i, (topic_num, score) in enumerate(zip(topic_nums, topic_scores)):
            if topic_num in self.topics:
                results.append((topic_num, self.topics[topic_num], float(score)))
                
        return results
    
    def search_documents(
        self,
        query: str,
        n_docs: int = 10,
    ) -> List[Tuple[str, float]]:
        """Search for documents similar to a query.
        
        Parameters
        ----------
        query : str
            Query string to search for
        n_docs : int, optional
            Number of documents to return, by default 10
            
        Returns
        -------
        List[Tuple[str, float]]
            List of tuples (document, score)
        """
        if not self.is_fitted:
            raise ValueError("Model must be fitted before searching documents")
            
        # Use Top2Vec's search_documents method
        doc_ids, doc_scores = self.model.search_documents_by_keywords(
            keywords=[query],
            num_docs=n_docs
        )
        
        # Get the documents
        documents = self.model.documents[doc_ids]
        
        # Create the result list
        results = []
        for doc, score in zip(documents, doc_scores):
            results.append((doc, float(score)))
            
        return results
    
    def save(self, path: Union[str, Path]) -> None:
        """Save the model to disk.
        
        Parameters
        ----------
        path : Union[str, Path]
            Path to save the model to
        """
        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)
        
        # Save Top2Vec model
        self.model.save(str(path / "top2vec_model"))
        
        # Save other attributes
        with open(path / "metadata.pkl", "wb") as f:
            pickle.dump({
                "n_topics": self.n_topics,
                "min_topic_size": self.min_topic_size,
                "use_gpu": self.use_gpu,
                "umap_args": self.umap_args,
                "hdbscan_args": self.hdbscan_args,
                "use_custom_embeddings": self.use_custom_embeddings,
                "verbose": self.verbose,
                "topics": self.topics,
                "topic_sizes": self.topic_sizes,
                "is_fitted": self.is_fitted,
            }, f)
    
    @classmethod
    def load(
        cls,
        path: Union[str, Path],
        embedding_model: Optional[DocumentEmbedding] = None,
    ) -> "Top2VecModel":
        """Load a model from disk.
        
        Parameters
        ----------
        path : Union[str, Path]
            Path to load the model from
        embedding_model : Optional[DocumentEmbedding], optional
            Document embedding model to use, by default None
            
        Returns
        -------
        Top2VecModel
            Loaded model
        """
        if not TOP2VEC_AVAILABLE:
            raise ImportError(
                "Top2Vec is required for this model. "
                "Install with 'pip install top2vec>=1.0.27'"
            )
            
        path = Path(path)
        
        # Load metadata
        with open(path / "metadata.pkl", "rb") as f:
            metadata = pickle.load(f)
            
        # Create instance with loaded parameters
        instance = cls(
            embedding_model=embedding_model,
            n_topics=metadata["n_topics"],
            min_topic_size=metadata["min_topic_size"],
            use_gpu=metadata["use_gpu"],
            umap_args=metadata["umap_args"],
            hdbscan_args=metadata["hdbscan_args"],
            use_custom_embeddings=metadata["use_custom_embeddings"],
            verbose=metadata["verbose"],
        )
        
        # Load Top2Vec model
        instance.model = Top2Vec.load(str(path / "top2vec_model"))
        
        # Load other attributes
        instance.topics = metadata["topics"]
        instance.topic_sizes = metadata["topic_sizes"]
        instance.is_fitted = metadata["is_fitted"]
        
        return instance


"""Tests for BERTopic model implementation."""

import pytest
import numpy as np
import pandas as pd
from typing import List
import os
from pathlib import Path
import tempfile

try:
    from bertopic import BERTopic
    from meno.modeling.bertopic_model import BERTopicModel
    from meno.modeling.embeddings import DocumentEmbedding
    BERTOPIC_AVAILABLE = True
except ImportError:
    BERTOPIC_AVAILABLE = False

# Skip these tests if bertopic is not available
pytestmark = pytest.mark.skipif(
    not BERTOPIC_AVAILABLE,
    reason="BERTopic dependencies not available"
)


class TestBERTopicModel:
    """Tests for the BERTopic model implementation."""

    @pytest.fixture
    def sample_texts(self) -> List[str]:
        """Sample texts for testing."""
        return [
            "This is a document about technology and computers.",
            "Healthcare and medicine are important for public health.",
            "Sports and exercise are good for physical health.",
            "Education and learning are lifelong pursuits.",
            "Politics and government policies affect many aspects of life.",
            "Art and creativity help express human emotions.",
            "Science advances our understanding of the world.",
            "History teaches us about past events and their impact.",
            "Economics studies how resources are allocated.",
            "Philosophy explores fundamental questions about existence."
        ] * 3  # Repeat to have enough documents for BERTopic to find patterns

    @pytest.fixture
    def embedding_model(self):
        """Create a test embedding model."""
        return DocumentEmbedding(
            model_name="all-MiniLM-L6-v2",
            device="cpu",
            batch_size=8
        )

    @pytest.fixture
    def bertopic_model(self, embedding_model):
        """Create a BERTopic model instance."""
        return BERTopicModel(
            embedding_model=embedding_model,
            n_topics=3,  # Small number for testing
            min_topic_size=2,
            use_gpu=False,
            n_neighbors=5,
            n_components=2,
            verbose=False
        )

    def test_model_initialization(self, bertopic_model):
        """Test that the model initializes correctly."""
        # Check attributes
        assert bertopic_model.n_topics == 3
        assert bertopic_model.min_topic_size == 2
        assert bertopic_model.use_gpu is False
        assert bertopic_model.is_fitted is False
        
        # Check components
        assert hasattr(bertopic_model, "model")
        assert isinstance(bertopic_model.model, BERTopic)
        assert hasattr(bertopic_model, "embedding_model")
        assert isinstance(bertopic_model.embedding_model, DocumentEmbedding)

    def test_fit_and_transform(self, bertopic_model, sample_texts):
        """Test fitting the model and transforming documents."""
        # Fit the model
        fitted_model = bertopic_model.fit(sample_texts)
        
        # Check it's fitted
        assert fitted_model.is_fitted is True
        assert len(fitted_model.topics) > 0
        assert len(fitted_model.topic_sizes) > 0
        
        # Check topic embeddings
        assert fitted_model.topic_embeddings is not None
        assert fitted_model.topic_embeddings.shape[1] == 384  # MiniLM dimension
        
        # Transform documents
        topics, probs = fitted_model.transform(sample_texts[:5])
        
        # Check outputs
        assert len(topics) == 5
        assert probs.shape[0] == 5

    def test_fit_with_embeddings(self, bertopic_model, embedding_model, sample_texts):
        """Test fitting with pre-computed embeddings."""
        # Generate embeddings
        embeddings = embedding_model.embed_documents(sample_texts)
        
        # Fit with pre-computed embeddings
        fitted_model = bertopic_model.fit(sample_texts, embeddings=embeddings)
        
        # Check it's fitted
        assert fitted_model.is_fitted is True
        assert len(fitted_model.topics) > 0
        
        # Transform with embeddings
        topics, probs = fitted_model.transform(sample_texts[:5], embeddings=embeddings[:5])
        
        # Check outputs
        assert len(topics) == 5
        assert probs.shape[0] == 5

    def test_find_similar_topics(self, bertopic_model, sample_texts):
        """Test finding similar topics."""
        # Fit the model
        fitted_model = bertopic_model.fit(sample_texts)
        
        # Find similar topics
        similar_topics = fitted_model.find_similar_topics("technology and computers", n_topics=2)
        
        # Check results
        assert len(similar_topics) <= 2  # May be fewer than 2 if not enough topics
        if len(similar_topics) > 0:
            # Each result should be a tuple of (topic_id, description, score)
            assert len(similar_topics[0]) == 3
            assert isinstance(similar_topics[0][0], int)  # topic_id
            assert isinstance(similar_topics[0][1], str)  # description
            assert isinstance(similar_topics[0][2], float)  # score

    def test_save_and_load(self, bertopic_model, sample_texts, embedding_model, tmp_path):
        """Test saving and loading the model."""
        # Fit the model
        fitted_model = bertopic_model.fit(sample_texts)
        
        # Save the model
        save_path = tmp_path / "bertopic_test"
        fitted_model.save(save_path)
        
        # Check that files were created
        assert (save_path / "metadata.json").exists()
        assert (save_path / "topic_embeddings.npy").exists()
        
        # Load the model
        loaded_model = BERTopicModel.load(save_path, embedding_model=embedding_model)
        
        # Check attributes
        assert loaded_model.n_topics == fitted_model.n_topics
        assert loaded_model.min_topic_size == fitted_model.min_topic_size
        assert loaded_model.is_fitted is True
        
        # Check topics
        assert set(loaded_model.topics.keys()) == set(fitted_model.topics.keys())
        
        # Make predictions with loaded model
        topics, probs = loaded_model.transform(sample_texts[:3])
        assert len(topics) == 3

    def test_visualizations(self, bertopic_model, sample_texts):
        """Test visualization methods."""
        # Fit the model
        fitted_model = bertopic_model.fit(sample_texts)
        
        # Get visualizations
        topics_viz = fitted_model.visualize_topics()
        hierarchy_viz = fitted_model.visualize_hierarchy()
        
        # Check that visualizations were created
        assert topics_viz is not None
        assert hierarchy_viz is not None
        
        # Basic checks on the visualization objects
        assert hasattr(topics_viz, "data")
        assert hasattr(hierarchy_viz, "data")


"""HTML report generation utilities."""

from typing import Dict, List, Optional, Union, Any
import os
import json
import pandas as pd
import numpy as np
from pathlib import Path
import jinja2
import plotly.graph_objects as go

from ..visualization.static_plots import plot_topic_distribution, plot_topic_word_clouds
from ..visualization.interactive_plots import (
    plot_embeddings, 
    plot_topic_clusters,
    plot_topic_similarity_heatmap,
    plot_interactive_wordcloud
)


# Enhanced HTML template for reports with modern styling and interactive features
DEFAULT_TEMPLATE = """
<!DOCTYPE html>
<html>
<head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <title>{{ title }}</title>
    <style>
        :root {
            --primary-color: #3498db;
            --secondary-color: #2c3e50;
            --accent-color: #9b59b6;
            --light-bg: #f8f9fa;
            --dark-bg: #2c3e50;
            --text-color: #333;
            --light-text: #f8f9fa;
            --border-color: #e9ecef;
            --topic-colors: #e74c3c, #3498db, #2ecc71, #f39c12, #9b59b6, #1abc9c;
        }
        
        * {
            box-sizing: border-box;
        }
        
        body {
            font-family: 'Segoe UI', Roboto, -apple-system, BlinkMacSystemFont, sans-serif;
            line-height: 1.6;
            margin: 0;
            padding: 0;
            color: var(--text-color);
            background-color: #fff;
        }
        
        .container {
            max-width: 1200px;
            margin: 0 auto;
            padding: 0 20px;
        }
        
        header {
            background-color: var(--dark-bg);
            color: var(--light-text);
            padding: 20px 0;
            margin-bottom: 30px;
        }
        
        header .container {
            display: flex;
            justify-content: space-between;
            align-items: center;
        }
        
        .header-meta {
            font-size: 0.9rem;
            opacity: 0.8;
        }
        
        h1, h2, h3, h4 {
            color: var(--secondary-color);
            font-weight: 600;
            margin-top: 0;
        }
        
        h1 {
            font-size: 2.2rem;
            color: var(--light-text);
            margin: 0;
        }
        
        h2 {
            font-size: 1.8rem;
            border-bottom: 2px solid var(--border-color);
            padding-bottom: 10px;
            margin-bottom: 20px;
        }
        
        h3 {
            font-size: 1.4rem;
        }
        
        .card {
            background-color: #fff;
            border-radius: 8px;
            box-shadow: 0 2px 10px rgba(0, 0, 0, 0.1);
            margin-bottom: 30px;
            overflow: hidden;
        }
        
        .card-header {
            background-color: var(--light-bg);
            padding: 15px 20px;
            border-bottom: 1px solid var(--border-color);
        }
        
        .card-body {
            padding: 20px;
        }
        
        .visualization {
            margin: 20px 0;
            width: 100%;
            overflow: hidden;
        }
        
        .topic-grid {
            display: grid;
            grid-template-columns: repeat(auto-fill, minmax(300px, 1fr));
            gap: 20px;
            margin-bottom: 30px;
        }
        
        .topic-card {
            background-color: #fff;
            border-radius: 8px;
            box-shadow: 0 2px 8px rgba(0, 0, 0, 0.1);
            overflow: hidden;
            transition: transform 0.2s ease-in-out, box-shadow 0.2s ease-in-out;
        }
        
        .topic-card:hover {
            transform: translateY(-5px);
            box-shadow: 0 5px 15px rgba(0, 0, 0, 0.1);
        }
        
        .topic-header {
            padding: 15px;
            border-bottom: 1px solid var(--border-color);
            display: flex;
            justify-content: space-between;
            align-items: center;
        }
        
        .topic-name {
            font-weight: 600;
            font-size: 1.2rem;
            margin: 0;
        }
        
        .topic-count {
            background-color: var(--primary-color);
            color: white;
            border-radius: 20px;
            padding: 3px 10px;
            font-size: 0.8rem;
            font-weight: 600;
        }
        
        .topic-body {
            padding: 15px;
        }
        
        .topic-words {
            background-color: var(--light-bg);
            padding: 8px 12px;
            border-radius: 4px;
            margin-bottom: 15px;
            font-size: 0.9rem;
            color: var(--secondary-color);
        }
        
        .example-text {
            background-color: var(--light-bg);
            padding: 10px 15px;
            border-left: 3px solid var(--primary-color);
            margin-bottom: 10px;
            border-radius: 4px;
            font-size: 0.9rem;
        }
        
        .tabs {
            display: flex;
            border-bottom: 1px solid var(--border-color);
            margin-bottom: 20px;
        }
        
        .tab {
            padding: 10px 20px;
            cursor: pointer;
            font-weight: 600;
            color: var(--text-color);
            opacity: 0.7;
            transition: opacity 0.2s;
        }
        
        .tab.active {
            opacity: 1;
            border-bottom: 3px solid var(--primary-color);
        }
        
        .tab-content {
            display: none;
        }
        
        .tab-content.active {
            display: block;
        }
        
        table {
            border-collapse: collapse;
            width: 100%;
            margin-bottom: 20px;
            font-size: 0.9rem;
        }
        
        th, td {
            padding: 12px 15px;
            text-align: left;
            border-bottom: 1px solid var(--border-color);
        }
        
        th {
            background-color: var(--light-bg);
            font-weight: 600;
        }
        
        tr:hover {
            background-color: #f5f5f5;
        }
        
        .summary-cards {
            display: grid;
            grid-template-columns: repeat(auto-fill, minmax(200px, 1fr));
            gap: 20px;
            margin-bottom: 30px;
        }
        
        .summary-card {
            background-color: var(--light-bg);
            border-radius: 8px;
            padding: 20px;
            text-align: center;
        }
        
        .summary-value {
            font-size: 2rem;
            font-weight: 700;
            color: var(--primary-color);
        }
        
        .summary-label {
            font-size: 0.9rem;
            color: var(--secondary-color);
            margin-top: 5px;
        }
        
        .export-btn {
            background-color: var(--primary-color);
            color: white;
            border: none;
            padding: 8px 16px;
            border-radius: 4px;
            cursor: pointer;
            font-weight: 600;
            transition: background-color 0.2s;
        }
        
        .export-btn:hover {
            background-color: #2980b9;
        }
        
        .sample-note {
            background-color: #f8f9fa;
            border-left: 3px solid var(--primary-color);
            padding: 10px 15px;
            margin-bottom: 15px;
            font-size: 0.9rem;
            color: var(--secondary-color);
            border-radius: 4px;
        }
        
        footer {
            background-color: var(--light-bg);
            padding: 20px 0;
            text-align: center;
            margin-top: 50px;
            font-size: 0.9rem;
            color: var(--secondary-color);
        }
        
        @media (max-width: 768px) {
            .topic-grid {
                grid-template-columns: 1fr;
            }
            
            .summary-cards {
                grid-template-columns: 1fr 1fr;
            }
        }
    </style>
    {{ plotly_js }}
    <script>
        // Add interactive functionality after document loads
        document.addEventListener('DOMContentLoaded', function() {
            // Tab functionality
            const tabs = document.querySelectorAll('.tab');
            const tabContents = document.querySelectorAll('.tab-content');
            
            tabs.forEach(tab => {
                tab.addEventListener('click', () => {
                    // Remove active class from all tabs and contents
                    tabs.forEach(t => t.classList.remove('active'));
                    tabContents.forEach(c => c.classList.remove('active'));
                    
                    // Add active class to clicked tab and corresponding content
                    tab.classList.add('active');
                    const contentId = tab.getAttribute('data-tab');
                    document.getElementById(contentId).classList.add('active');
                });
            });
            
            // Full CSV data for export
            const fullData = {{ full_data_json|safe }};
            
            // Export table to CSV
            const exportBtn = document.getElementById('export-csv');
            if (exportBtn) {
                exportBtn.addEventListener('click', () => {
                    if (!fullData || !fullData.columns || !fullData.data) {
                        // Fall back to exporting visible table if full data isn't available
                        exportVisibleTable();
                        return;
                    }
                    
                    // Create CSV from full dataset
                    let csv = [];
                    
                    // Add header row
                    csv.push(fullData.columns.map(col => `"${col}"`).join(','));
                    
                    // Add data rows
                    fullData.data.forEach(row => {
                        const rowData = row.map(cell => {
                            // Handle null values
                            if (cell === null) return '""';
                            // Format numbers to 3 decimal places
                            if (typeof cell === 'number') {
                                return `"${cell.toFixed(3)}"`;
                            }
                            // Escape quotes in text
                            return `"${String(cell).replace(/"/g, '""')}"`;
                        });
                        csv.push(rowData.join(','));
                    });
                    
                    downloadCSV(csv.join('\\n'));
                });
            }
            
            // Function to export only the visible table
            function exportVisibleTable() {
                const table = document.querySelector('.data-table');
                if (!table) return;
                
                let csv = [];
                const rows = table.querySelectorAll('tr');
                
                rows.forEach(row => {
                    const cols = row.querySelectorAll('td, th');
                    const rowData = Array.from(cols).map(col => '"' + col.innerText.replace(/"/g, '""') + '"');
                    csv.push(rowData.join(','));
                });
                
                downloadCSV(csv.join('\\n'));
            }
            
            // Helper function to download CSV data
            function downloadCSV(csvContent) {
                const blob = new Blob([csvContent], { type: 'text/csv' });
                const url = URL.createObjectURL(blob);
                
                const a = document.createElement('a');
                a.setAttribute('hidden', '');
                a.setAttribute('href', url);
                a.setAttribute('download', 'topic_data.csv');
                document.body.appendChild(a);
                a.click();
                document.body.removeChild(a);
            }
        });
    </script>
</head>
<body>
    <header>
        <div class="container">
            <h1>{{ title }}</h1>
            <div class="header-meta">
                Generated on {{ generation_date }}
            </div>
        </div>
    </header>

    <div class="container">
        <div class="card">
            <div class="card-header">
                <h2>Summary</h2>
            </div>
            <div class="card-body">
                <div class="summary-cards">
                    <div class="summary-card">
                        <div class="summary-value">{{ document_count }}</div>
                        <div class="summary-label">Documents</div>
                    </div>
                    <div class="summary-card">
                        <div class="summary-value">{{ topic_count }}</div>
                        <div class="summary-label">Topics</div>
                    </div>
                    {% if avg_words_per_doc %}
                    <div class="summary-card">
                        <div class="summary-value">{{ avg_words_per_doc }}</div>
                        <div class="summary-label">Avg Words/Doc</div>
                    </div>
                    {% endif %}
                    {% if topic_coherence %}
                    <div class="summary-card">
                        <div class="summary-value">{{ topic_coherence }}</div>
                        <div class="summary-label">Topic Coherence</div>
                    </div>
                    {% endif %}
                </div>
            </div>
        </div>

        <div class="card">
            <div class="card-header">
                <h2>Topic Distribution</h2>
            </div>
            <div class="card-body">
                <div class="visualization">
                    {{ topic_distribution_plot }}
                </div>
            </div>
        </div>

        {% if umap_plot %}
        <div class="card">
            <div class="card-header">
                <h2>Visualizations</h2>
            </div>
            <div class="card-body">
                <div class="tabs">
                    <div class="tab active" data-tab="embeddings-tab">Document Embeddings</div>
                    {% if topic_similarity_plot %}
                    <div class="tab" data-tab="similarity-tab">Topic Similarity</div>
                    {% endif %}
                    {% if topic_wordcloud_plot %}
                    <div class="tab" data-tab="wordcloud-tab">Word Clouds</div>
                    {% endif %}
                </div>
                
                <div id="embeddings-tab" class="tab-content active">
                    <div class="visualization">
                        {{ umap_plot }}
                    </div>
                </div>
                
                {% if topic_similarity_plot %}
                <div id="similarity-tab" class="tab-content">
                    <div class="visualization">
                        {{ topic_similarity_plot }}
                    </div>
                </div>
                {% endif %}
                
                {% if topic_wordcloud_plot %}
                <div id="wordcloud-tab" class="tab-content">
                    <div class="visualization">
                        {{ topic_wordcloud_plot }}
                    </div>
                </div>
                {% endif %}
            </div>
        </div>
        {% endif %}

        <div class="card">
            <div class="card-header">
                <h2>Topic Details</h2>
            </div>
            <div class="card-body">
                <div class="topic-grid">
                    {% for topic in topics %}
                    <div class="topic-card">
                        <div class="topic-header">
                            <h3 class="topic-name">{{ topic.name }}</h3>
                            <span class="topic-count">{{ topic.count }}</span>
                        </div>
                        <div class="topic-body">
                            {% if topic.top_words %}
                            <div class="topic-words">
                                <strong>Top words:</strong> {{ topic.top_words }}
                            </div>
                            {% endif %}
                            
                            <h4>Example Documents:</h4>
                            {% for example in topic.examples %}
                            <div class="example-text">{{ example }}</div>
                            {% endfor %}
                        </div>
                    </div>
                    {% endfor %}
                </div>
            </div>
        </div>

        {% if include_raw_data %}
        <div class="card">
            <div class="card-header">
                <div style="display: flex; justify-content: space-between; align-items: center;">
                    <h2 style="margin: 0;">Raw Data</h2>
                    <button id="export-csv" class="export-btn">Export to CSV</button>
                </div>
            </div>
            <div class="card-body">
                <div id="topic-assignments-table">
                    {{ topic_assignments_table }}
                </div>
            </div>
        </div>
        {% endif %}
    </div>
    
    <footer>
        <div class="container">
            Generated with Meno Topic Modeling Toolkit
        </div>
    </footer>
</body>
</html>
"""


def generate_html_report(
    documents: pd.DataFrame,
    topic_assignments: pd.DataFrame,
    umap_projection: Optional[np.ndarray] = None,
    output_path: Optional[Union[str, Path]] = None,
    config: Optional[Union[Dict[str, Any], "HTMLReportConfig"]] = None,
    template: Optional[str] = None,
    similarity_matrix: Optional[np.ndarray] = None,
    topic_words: Optional[Dict[str, Dict[str, float]]] = None,
) -> str:
    """Generate an HTML report of topic modeling results.
    
    Parameters
    ----------
    documents : pd.DataFrame
        DataFrame with document texts and topic assignments
    topic_assignments : pd.DataFrame
        DataFrame with topic assignments and probabilities
    umap_projection : Optional[np.ndarray], optional
        UMAP projection for visualization, by default None
    output_path : Optional[Union[str, Path]], optional
        Path to save the report, by default None
        If None, creates a file in the current directory
    config : Optional[Union[Dict[str, Any], HTMLReportConfig]], optional
        Report configuration, by default None
    template : Optional[str], optional
        Custom Jinja2 template, by default None
    similarity_matrix : Optional[np.ndarray], optional
        Matrix of topic similarities, shape (n_topics, n_topics), by default None
    topic_words : Optional[Dict[str, Dict[str, float]]], optional
        Dictionary mapping topic names to word frequency dictionaries, by default None
    
    Returns
    -------
    str
        Path to the generated report
    """
    # Default configuration
    default_config = {
        "title": "Topic Modeling Results",
        "include_interactive": True,
        "max_examples_per_topic": 5,
        "include_raw_data": False,
    }
    
    # Handle Pydantic model and convert to dict if needed
    if config is not None and hasattr(config, "dict"):
        config = config.dict()
    
    # Merge with provided config
    config = {**default_config, **(config or {})}
    
    # Use default template if not provided
    if template is None:
        template = DEFAULT_TEMPLATE
    
    # Create output path if not provided
    if output_path is None:
        from datetime import datetime
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        output_path = f"topic_modeling_report_{timestamp}.html"
    
    # Convert to Path object
    output_path = Path(output_path)
    
    # Create plotly JS header for interactive visualizations
    plotly_js = ""
    if config["include_interactive"]:
        plotly_js = '<script src="https://cdn.plot.ly/plotly-latest.min.js"></script>'
    
    # Create topic distribution plot
    topic_distribution_fig = plot_topic_distribution(
        documents["topic"],
        title="Topic Distribution",
        sort_by_count=True,
    )
    topic_distribution_plot = topic_distribution_fig.to_html(
        full_html=False,
        include_plotlyjs=False,
    )
    
    # Create UMAP plot if projection provided
    umap_plot = None
    if umap_projection is not None and config["include_interactive"]:
        umap_fig = plot_embeddings(
            umap_projection,
            documents["topic"],
            document_texts=documents["text"],
            title="Document Embeddings",
        )
        umap_plot = umap_fig.to_html(
            full_html=False,
            include_plotlyjs=False,
        )
    
    # Create topic similarity heatmap if similarity matrix provided
    topic_similarity_plot = None
    if similarity_matrix is not None and config["include_interactive"]:
        topic_names = sorted(documents["topic"].unique())
        similarity_fig = plot_topic_similarity_heatmap(
            similarity_matrix,
            topic_names,
            title="Topic Similarity",
        )
        topic_similarity_plot = similarity_fig.to_html(
            full_html=False,
            include_plotlyjs=False,
        )
    
    # Create interactive word cloud if topic_words provided
    topic_wordcloud_plot = None
    if topic_words is not None and config["include_interactive"]:
        wordcloud_fig = plot_interactive_wordcloud(
            topic_words,
            title="Topic Word Clouds",
        )
        topic_wordcloud_plot = wordcloud_fig.to_html(
            full_html=False,
            include_plotlyjs=False,
        )
    
    # Get topic information
    topics_info = []
    for topic_name, topic_group in documents.groupby("topic"):
        # Skip if no documents
        if len(topic_group) == 0:
            continue
        
        # Get top words if available
        top_words = ""
        if "top_words" in topic_assignments.columns:
            top_words_row = topic_assignments[
                topic_assignments["topic"] == topic_name
            ]["top_words"].iloc[0]
            if isinstance(top_words_row, str):
                top_words = top_words_row
        elif topic_words and topic_name in topic_words:
            # Use top words from topic_words if available
            words = topic_words[topic_name]
            top_words = ", ".join(
                sorted(words.keys(), key=lambda k: words[k], reverse=True)[:10]
            )
        
        # Get example documents
        examples = topic_group["text"].head(config["max_examples_per_topic"]).tolist()
        
        # Add topic info
        topics_info.append({
            "name": topic_name,
            "count": len(topic_group),
            "top_words": top_words,
            "examples": examples,
        })
    
    # Sort topics by count
    topics_info.sort(key=lambda x: x["count"], reverse=True)
    
    # Create HTML table for raw data
    topic_assignments_table = ""
    full_data_json = "{}"
    if config["include_raw_data"]:
        # Select subset of columns for display
        display_cols = ["topic", "topic_probability"]
        
        # Add topic similarity columns if available
        similarity_cols = [
            col for col in topic_assignments.columns
            if col.endswith("_similarity")
        ]
        display_cols.extend(similarity_cols)
        
        # Create a sample of the data by taking a few examples from each topic
        max_samples_per_topic = config.get("max_samples_per_topic", 5)
        
        # Get unique topics
        unique_topics = documents["topic"].unique()
        
        # Create sample DataFrame with examples from each topic
        sample_rows = []
        for topic_name in unique_topics:
            # Get rows for this topic, limited to max_samples_per_topic
            topic_rows = documents[documents["topic"] == topic_name].head(max_samples_per_topic)
            sample_rows.append(topic_rows)
        
        # Combine all sample rows
        sample_df = pd.concat(sample_rows)
        
        # Create HTML table
        table_df = pd.concat(
            [sample_df[["text"]], topic_assignments.loc[sample_df.index][display_cols]],
            axis=1,
        )
        
        # Add a note about the sampling
        sample_note = f"<div class='sample-note'><em>Note: Showing sample data ({max_samples_per_topic} examples per topic). Download CSV for full dataset.</em></div>"
        
        # Generate the HTML table
        topic_assignments_table = sample_note + table_df.to_html(
            index=False,
            classes="dataframe data-table",  # Add data-table class for CSV export
            float_format="%.3f",
        )
        
        # Create full data for CSV export
        full_table_df = pd.concat(
            [documents[["text"]], topic_assignments[display_cols]],
            axis=1,
        )
        
        # Convert to JSON for embedding in the HTML
        import json
        full_data_dict = {
            "columns": full_table_df.columns.tolist(),
            "data": full_table_df.values.tolist(),
        }
        full_data_json = json.dumps(full_data_dict)
    
    # Calculate additional metrics
    avg_words_per_doc = None
    if "text" in documents.columns:
        # Calculate average words per document
        avg_words = documents["text"].str.split().str.len().mean()
        avg_words_per_doc = f"{avg_words:.1f}"
    
    # Topic coherence score (placeholder - would be calculated by modeling)
    topic_coherence = None
    if "coherence" in topic_assignments.columns:
        topic_coherence = f"{topic_assignments['coherence'].mean():.2f}"
    
    # Create context for template
    from datetime import datetime
    context = {
        "title": config["title"],
        "document_count": len(documents),
        "topic_count": len(topics_info),
        "generation_date": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
        "plotly_js": plotly_js,
        "topic_distribution_plot": topic_distribution_plot,
        "umap_plot": umap_plot,
        "topic_similarity_plot": topic_similarity_plot,
        "topic_wordcloud_plot": topic_wordcloud_plot,
        "topics": topics_info,
        "include_raw_data": config["include_raw_data"],
        "topic_assignments_table": topic_assignments_table,
        "avg_words_per_doc": avg_words_per_doc,
        "topic_coherence": topic_coherence,
        "full_data_json": full_data_json,
    }
    
    # Render template
    html = jinja2.Template(template).render(**context)
    
    # Write to file
    with open(output_path, "w", encoding="utf-8") as f:
        f.write(html)
    
    return str(output_path)


def generate_json_report(
    documents: pd.DataFrame,
    topic_assignments: pd.DataFrame,
    output_path: Optional[Union[str, Path]] = None,
) -> str:
    """Generate a JSON report of topic modeling results.
    
    Parameters
    ----------
    documents : pd.DataFrame
        DataFrame with document texts and topic assignments
    topic_assignments : pd.DataFrame
        DataFrame with topic assignments and probabilities
    output_path : Optional[Union[str, Path]], optional
        Path to save the report, by default None
        If None, creates a file in the current directory
    
    Returns
    -------
    str
        Path to the generated report
    """
    # Create output path if not provided
    if output_path is None:
        from datetime import datetime
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        output_path = f"topic_modeling_report_{timestamp}.json"
    
    # Convert to Path object
    output_path = Path(output_path)
    
    # Prepare data for JSON
    from datetime import datetime
    
    # Get topic information
    topics_info = []
    for topic_name, topic_group in documents.groupby("topic"):
        # Get example documents
        examples = topic_group["text"].head(5).tolist()
        
        # Add topic info
        topics_info.append({
            "name": topic_name,
            "count": int(len(topic_group)),
            "examples": examples,
        })
    
    # Create report data
    report_data = {
        "metadata": {
            "document_count": len(documents),
            "topic_count": len(topics_info),
            "generation_date": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
        },
        "topics": topics_info,
        "topic_distribution": documents["topic"].value_counts().to_dict(),
    }
    
    # Write to file
    with open(output_path, "w", encoding="utf-8") as f:
        json.dump(report_data, f, indent=2)
    
    return str(output_path)


"""BERTopic model for topic modeling."""

from typing import List, Dict, Optional, Union, Any, Tuple
import numpy as np
import pandas as pd
import logging
from pathlib import Path
import json
import pickle

try:
    from bertopic import BERTopic
    from bertopic.representation import KeyBERTInspired
    from bertopic.vectorizers import ClassTfidfTransformer
    from bertopic.dimensionality import UMAPReducer
    BERTOPIC_AVAILABLE = True
except ImportError:
    BERTOPIC_AVAILABLE = False

from meno.modeling.embeddings import DocumentEmbedding
from meno.modeling.base import BaseTopicModel


class BERTopicModel(BaseTopicModel):
    """BERTopic model for topic modeling using BERT embeddings with UMAP and HDBSCAN.
    
    Parameters
    ----------
    n_topics : Optional[int], optional
        Number of topics to extract, by default None
        If None, BERTopic automatically determines the number of topics
    embedding_model : Optional[DocumentEmbedding], optional
        Document embedding model to use, by default None
        If None, a default DocumentEmbedding will be created
    min_topic_size : int, optional
        Minimum size of topics, by default 10
    use_gpu : bool, optional
        Whether to use GPU acceleration if available, by default False
        Setting to False (default) ensures CPU-only operation and avoids CUDA dependencies
    n_neighbors : int, optional
        Number of neighbors for UMAP, by default 15
    n_components : int, optional
        Number of dimensions for UMAP, by default 5
    verbose : bool, optional
        Whether to show verbose output, by default True
    
    Attributes
    ----------
    model : BERTopic
        Trained BERTopic model
    topics : Dict[int, str]
        Mapping of topic IDs to topic descriptions
    topic_sizes : Dict[int, int]
        Mapping of topic IDs to topic sizes
    """
    
    def __init__(
        self,
        n_topics: Optional[int] = None,
        embedding_model: Optional[DocumentEmbedding] = None,
        min_topic_size: int = 10,
        use_gpu: bool = False,
        n_neighbors: int = 15,
        n_components: int = 5,
        verbose: bool = True,
    ):
        """Initialize the BERTopic model."""
        if not BERTOPIC_AVAILABLE:
            raise ImportError(
                "BERTopic is required for this model. "
                "Install with 'pip install bertopic>=0.15.0'"
            )
            
        self.n_topics = n_topics
        self.min_topic_size = min_topic_size
        self.use_gpu = use_gpu
        self.n_neighbors = n_neighbors
        self.n_components = n_components
        self.verbose = verbose
        
        # Set up embedding model if not provided - default to CPU
        if embedding_model is None:
            self.embedding_model = DocumentEmbedding(use_gpu=False)  # Default to CPU
        else:
            self.embedding_model = embedding_model
            
        # Set up UMAP reducer
        self.umap_model = UMAPReducer(
            n_neighbors=n_neighbors,
            n_components=n_components,
            metric="cosine",
            low_memory=True,
        )
        
        # Set up representation model
        self.representation_model = KeyBERTInspired()
        
        # Set up vectorizer
        self.vectorizer_model = ClassTfidfTransformer()
        
        # Initialize BERTopic model
        self.model = BERTopic(
            nr_topics=n_topics,
            min_topic_size=min_topic_size,
            umap_model=self.umap_model,
            vectorizer_model=self.vectorizer_model,
            representation_model=self.representation_model,
            verbose=verbose,
        )
        
        # Initialize empty attributes
        self.topics = {}
        self.topic_sizes = {}
        self.topic_embeddings = None
        self.is_fitted = False
    
    def fit(
        self,
        documents: Union[List[str], pd.Series],
        embeddings: Optional[np.ndarray] = None,
    ) -> "BERTopicModel":
        """Fit the BERTopic model to a set of documents.
        
        Parameters
        ----------
        documents : Union[List[str], pd.Series]
            List or Series of document texts
        embeddings : Optional[np.ndarray], optional
            Pre-computed document embeddings, by default None
            If None, embeddings will be computed using the embedding model
        
        Returns
        -------
        BERTopicModel
            Fitted model
        """
        # Convert pandas Series to list if needed
        if isinstance(documents, pd.Series):
            documents = documents.tolist()
            
        # Generate embeddings if not provided
        if embeddings is None:
            embeddings = self.embedding_model.embed_documents(documents)
            
        # Fit BERTopic model
        topics, probs = self.model.fit_transform(documents, embeddings=embeddings)
        
        # Store topic information
        self.topics = {i: f"Topic {i}" for i in set(topics) if i != -1}
        self.topic_sizes = {
            topic: (topics == topic).sum() for topic in self.topics.keys()
        }
        
        # Add -1 for outliers
        if -1 in set(topics):
            self.topics[-1] = "Other"
            self.topic_sizes[-1] = (topics == -1).sum()
            
        # Update topic descriptions with top words
        topic_info = self.model.get_topic_info()
        for topic_id, row in topic_info.iterrows():
            if topic_id in self.topics and topic_id != -1:
                words = [word for word, _ in self.model.get_topic(topic_id)][:5]
                self.topics[topic_id] = f"Topic {topic_id}: {', '.join(words)}"
                
        # Compute topic embeddings
        self._compute_topic_embeddings()
        
        self.is_fitted = True
        return self
    
    def transform(
        self,
        documents: Union[List[str], pd.Series],
        embeddings: Optional[np.ndarray] = None,
    ) -> Tuple[List[int], np.ndarray]:
        """Transform documents to topic assignments and probabilities.
        
        Parameters
        ----------
        documents : Union[List[str], pd.Series]
            List or Series of document texts
        embeddings : Optional[np.ndarray], optional
            Pre-computed document embeddings, by default None
            If None, embeddings will be computed using the embedding model
        
        Returns
        -------
        Tuple[List[int], np.ndarray]
            Tuple of (topic_ids, probabilities)
        """
        if not self.is_fitted:
            raise ValueError("Model must be fitted before transform can be called")
            
        # Convert pandas Series to list if needed
        if isinstance(documents, pd.Series):
            documents = documents.tolist()
            
        # Generate embeddings if not provided
        if embeddings is None:
            embeddings = self.embedding_model.embed_documents(documents)
            
        # Transform documents
        topics, probs = self.model.transform(documents, embeddings=embeddings)
        
        return topics, probs
    
    def _compute_topic_embeddings(self) -> None:
        """Compute embeddings for all topics."""
        # Get topic descriptions with top words
        topic_info = self.model.get_topic_info()
        
        # Create descriptions for each topic
        descriptions = []
        topic_ids = []
        
        for topic_id, row in topic_info.iterrows():
            if topic_id != -1:  # Skip outlier topic
                words = [word for word, _ in self.model.get_topic(topic_id)][:10]
                description = " ".join(words)
                descriptions.append(description)
                topic_ids.append(topic_id)
                
        # Compute embeddings for descriptions
        if descriptions:
            self.topic_embeddings = self.embedding_model.embed_documents(descriptions)
            self.topic_id_mapping = {i: topic_id for i, topic_id in enumerate(topic_ids)}
    
    def find_similar_topics(
        self,
        query: str,
        n_topics: int = 5,
    ) -> List[Tuple[int, str, float]]:
        """Find topics similar to a query string.
        
        Parameters
        ----------
        query : str
            Query string to find similar topics for
        n_topics : int, optional
            Number of similar topics to return, by default 5
        
        Returns
        -------
        List[Tuple[int, str, float]]
            List of tuples (topic_id, topic_description, similarity_score)
        """
        if not self.is_fitted:
            raise ValueError("Model must be fitted before finding similar topics")
            
        if self.topic_embeddings is None:
            self._compute_topic_embeddings()
            
        if len(self.topic_embeddings) == 0:
            return []
            
        # Compute query embedding
        query_embedding = self.embedding_model.embed_documents([query])[0]
        
        # Compute similarities
        similarities = np.dot(self.topic_embeddings, query_embedding)
        
        # Get top n_topics
        top_indices = similarities.argsort()[-n_topics:][::-1]
        
        # Return topic IDs, descriptions, and similarity scores
        return [
            (
                self.topic_id_mapping[i],
                self.topics[self.topic_id_mapping[i]],
                float(similarities[i])
            )
            for i in top_indices
        ]
    
    def visualize_topics(
        self,
        width: int = 800,
        height: int = 600,
    ) -> Any:
        """Visualize topics using BERTopic's visualization tools.
        
        Parameters
        ----------
        width : int, optional
            Width of the visualization, by default 800
        height : int, optional
            Height of the visualization, by default 600
        
        Returns
        -------
        Any
            Plotly figure for topic visualization
        """
        if not self.is_fitted:
            raise ValueError("Model must be fitted before visualization")
            
        # Use BERTopic's topic visualization
        return self.model.visualize_topics(width=width, height=height)
    
    def visualize_hierarchy(
        self,
        width: int = 1000,
        height: int = 600,
    ) -> Any:
        """Visualize topic hierarchy.
        
        Parameters
        ----------
        width : int, optional
            Width of the visualization, by default 1000
        height : int, optional
            Height of the visualization, by default 600
        
        Returns
        -------
        Any
            Plotly figure for hierarchy visualization
        """
        if not self.is_fitted:
            raise ValueError("Model must be fitted before visualization")
            
        # Use BERTopic's hierarchical visualization
        return self.model.visualize_hierarchy(width=width, height=height)
    
    def save(self, path: Union[str, Path]) -> None:
        """Save the model to disk.
        
        Parameters
        ----------
        path : Union[str, Path]
            Path to save the model to
        """
        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)
        
        # Save BERTopic model
        self.model.save(path / "bertopic_model")
        
        # Save other attributes
        with open(path / "metadata.json", "w") as f:
            json.dump({
                "n_topics": self.n_topics,
                "min_topic_size": self.min_topic_size,
                "n_neighbors": self.n_neighbors,
                "n_components": self.n_components,
                "topics": {str(k): v for k, v in self.topics.items()},
                "topic_sizes": {str(k): v for k, v in self.topic_sizes.items()},
                "is_fitted": self.is_fitted,
                "topic_id_mapping": {str(k): v for k, v in self.topic_id_mapping.items()} if hasattr(self, "topic_id_mapping") else None,
            }, f)
            
        # Save topic embeddings
        if self.topic_embeddings is not None:
            np.save(path / "topic_embeddings.npy", self.topic_embeddings)
            
    @classmethod
    def load(
        cls,
        path: Union[str, Path],
        embedding_model: Optional[DocumentEmbedding] = None,
    ) -> "BERTopicModel":
        """Load a model from disk.
        
        Parameters
        ----------
        path : Union[str, Path]
            Path to load the model from
        embedding_model : Optional[DocumentEmbedding], optional
            Document embedding model to use, by default None
            
        Returns
        -------
        BERTopicModel
            Loaded model
        """
        if not BERTOPIC_AVAILABLE:
            raise ImportError(
                "BERTopic is required for this model. "
                "Install with 'pip install bertopic>=0.15.0'"
            )
            
        path = Path(path)
        
        # Load metadata
        with open(path / "metadata.json", "r") as f:
            metadata = json.load(f)
            
        # Create instance with loaded parameters
        instance = cls(
            n_topics=metadata["n_topics"],
            min_topic_size=metadata["min_topic_size"],
            n_neighbors=metadata["n_neighbors"],
            n_components=metadata["n_components"],
            embedding_model=embedding_model,
        )
        
        # Load BERTopic model
        instance.model = BERTopic.load(path / "bertopic_model")
        
        # Load other attributes
        instance.topics = {int(k): v for k, v in metadata["topics"].items()}
        instance.topic_sizes = {int(k): v for k, v in metadata["topic_sizes"].items()}
        instance.is_fitted = metadata["is_fitted"]
        
        if metadata["topic_id_mapping"] is not None:
            instance.topic_id_mapping = {int(k): v for k, v in metadata["topic_id_mapping"].items()}
        
        # Load topic embeddings if they exist
        topic_embeddings_path = path / "topic_embeddings.npy"
        if topic_embeddings_path.exists():
            instance.topic_embeddings = np.load(topic_embeddings_path)
            
        return instance


"""Document embedding module using transformer models (CPU-optimized by default)."""

from typing import List, Dict, Optional, Union, Any, Callable, Generator, Iterable
import os
import numpy as np
import pandas as pd
from sentence_transformers import SentenceTransformer
from transformers import AutoModel, AutoTokenizer
import torch

try:
    import polars as pl
    POLARS_AVAILABLE = True
except ImportError:
    POLARS_AVAILABLE = False


class DocumentEmbedding:
    """Generate and manage document embeddings using transformer models.
    
    Parameters
    ----------
    model_name : str, optional
        Name of the transformer model to use, by default "answerdotai/ModernBERT-base"
        Supports any model compatible with sentence-transformers or HuggingFace transformers.
    device : str, optional
        Device to run the model on, by default determined by use_gpu setting
    batch_size : int, optional
        Batch size for embedding generation, by default 32
    use_gpu : bool, optional
        Whether to use GPU acceleration if available, by default False
        Setting to False (default) ensures CPU-only operation and avoids CUDA dependencies
    local_model_path : str, optional
        Path to locally downloaded model, by default None (will download from HuggingFace)
    
    Attributes
    ----------
    model : SentenceTransformer
        Loaded transformer model
    embedding_dim : int
        Dimension of the embeddings
    device : str
        Device being used for inference (either "cpu" or "cuda")
    """
    
    def __init__(
        self,
        model_name: str = "answerdotai/ModernBERT-base",
        device: Optional[str] = None,
        batch_size: int = 32,
        use_gpu: bool = False,
        local_model_path: Optional[str] = None,
    ):
        """Initialize the document embedding model."""
        self.model_name = model_name
        self.batch_size = batch_size
        self.use_gpu = use_gpu
        self.local_model_path = local_model_path
        
        # Set device - default to CPU
        if device is None:
            # Only use CUDA if explicitly requested and available
            self.device = "cuda" if (torch.cuda.is_available() and use_gpu) else "cpu"
        else:
            self.device = device
            
        # Load model
        if local_model_path and os.path.exists(local_model_path):
            self.model = SentenceTransformer(local_model_path, device=self.device)
        else:
            self.model = SentenceTransformer(model_name, device=self.device)
            
        self.embedding_dim = self.model.get_sentence_embedding_dimension()
    
    def _create_batches(self, documents: List[str], batch_size: int) -> Generator[List[str], None, None]:
        """Split documents into batches for processing.
        
        Parameters
        ----------
        documents : List[str]
            List of document texts to split into batches
        batch_size : int
            Size of each batch
            
        Yields
        ------
        List[str]
            Batch of documents
        """
        for i in range(0, len(documents), batch_size):
            yield documents[i:i + batch_size]
    
    def embed_documents(
        self, 
        documents: Union[List[str], pd.Series, "pl.Series", "pl.DataFrame"],
        show_progress_bar: bool = True,
        text_column: Optional[str] = None,
    ) -> np.ndarray:
        """Generate embeddings for a list of documents.
        
        Parameters
        ----------
        documents : Union[List[str], pd.Series, pl.Series, pl.DataFrame]
            List, Series, or DataFrame of document texts to embed.
            If DataFrame, must provide text_column.
        show_progress_bar : bool, optional
            Whether to show a progress bar, by default True
        text_column : Optional[str], optional
            If documents is a DataFrame, the name of the column containing text, by default None
        
        Returns
        -------
        np.ndarray
            Document embeddings with shape (len(documents), embedding_dim)
        """
        # Handle different input types, converting to list
        if isinstance(documents, pd.Series):
            documents = documents.tolist()
        elif POLARS_AVAILABLE and isinstance(documents, pl.Series):
            documents = documents.to_list()
        elif POLARS_AVAILABLE and isinstance(documents, pl.DataFrame):
            if text_column is None:
                raise ValueError("text_column must be provided when documents is a DataFrame")
            documents = documents[text_column].to_list()
        elif isinstance(documents, pd.DataFrame):
            if text_column is None:
                raise ValueError("text_column must be provided when documents is a DataFrame")
            documents = documents[text_column].tolist()
        
        # Generate embeddings
        embeddings = self.model.encode(
            documents,
            batch_size=self.batch_size,
            show_progress_bar=show_progress_bar,
            convert_to_numpy=True,
        )
        
        return embeddings
    
    def embed_documents_stream(
        self,
        documents_iterator: Iterable[str],
        batch_size: Optional[int] = None,
    ) -> Generator[np.ndarray, None, None]:
        """Generate embeddings for a stream of documents, yielding batches of embeddings.
        
        This is useful for large datasets that don't fit in memory.
        
        Parameters
        ----------
        documents_iterator : Iterable[str]
            Iterator yielding document texts
        batch_size : Optional[int], optional
            Batch size for processing, by default None (uses self.batch_size)
        
        Yields
        ------
        np.ndarray
            Batches of document embeddings with shape (batch_size, embedding_dim)
        """
        if batch_size is None:
            batch_size = self.batch_size
            
        # Process documents in batches
        batch = []
        for doc in documents_iterator:
            batch.append(doc)
            if len(batch) >= batch_size:
                yield self.embed_documents(batch, show_progress_bar=False)
                batch = []
                
        # Process final batch if there are any remaining documents
        if batch:
            yield self.embed_documents(batch, show_progress_bar=False)
    
    def embed_polars_streaming(
        self,
        data_path: str,
        text_column: str,
        batch_size: Optional[int] = None,
        filter_condition: Optional[Callable] = None,
    ) -> Generator[np.ndarray, None, None]:
        """Stream a large dataset using Polars for memory efficiency.
        
        Parameters
        ----------
        data_path : str
            Path to the data file (CSV, Parquet, etc.)
        text_column : str
            Name of the column containing text to embed
        batch_size : Optional[int], optional
            Batch size for processing, by default None (uses self.batch_size)
        filter_condition : Optional[Callable], optional
            Function taking a Polars DataFrame and returning a filtered DataFrame,
            by default None
            
        Yields
        ------
        np.ndarray
            Batches of document embeddings
            
        Raises
        ------
        ImportError
            If Polars is not available
        """
        if not POLARS_AVAILABLE:
            raise ImportError(
                "Polars is required for streaming large datasets. "
                "Install with 'pip install \"meno[optimization]\"'"
            )
            
        if batch_size is None:
            batch_size = self.batch_size
        
        # Determine file format from extension
        file_format = os.path.splitext(data_path)[1].lower()
        
        # Use Polars lazy loading and streaming
        if file_format == ".csv":
            reader = pl.scan_csv(data_path)
        elif file_format in (".parquet", ".pq"):
            reader = pl.scan_parquet(data_path)
        else:
            raise ValueError(f"Unsupported file format: {file_format}")
        
        # Apply filter if provided
        if filter_condition is not None:
            reader = filter_condition(reader)
        
        # Set up streaming with batches
        streaming_df = reader.select(text_column).collect(streaming=True)
        
        # Process in batches using the streaming iterator
        for batch_df in streaming_df.iter_batches(batch_size=batch_size):
            texts = batch_df[text_column].to_list()
            yield self.embed_documents(texts, show_progress_bar=False)
    
    def embed_topics(
        self,
        topics: Union[List[str], pd.Series],
        descriptions: Optional[Union[List[str], pd.Series]] = None,
    ) -> np.ndarray:
        """Generate embeddings for a list of topics, optionally with descriptions.
        
        Parameters
        ----------
        topics : Union[List[str], pd.Series]
            List or Series of topic names
        descriptions : Optional[Union[List[str], pd.Series]], optional
            List or Series of topic descriptions, by default None
            If provided, topics and descriptions are combined
        
        Returns
        -------
        np.ndarray
            Topic embeddings with shape (len(topics), embedding_dim)
        """
        # Convert to lists if needed
        if isinstance(topics, pd.Series):
            topics = topics.tolist()
        
        if descriptions is not None:
            if isinstance(descriptions, pd.Series):
                descriptions = descriptions.tolist()
            
            # Combine topics and descriptions
            texts_to_embed = [
                f"{topic}: {desc}" for topic, desc in zip(topics, descriptions)
            ]
        else:
            texts_to_embed = topics
        
        # Generate embeddings
        return self.embed_documents(texts_to_embed, show_progress_bar=False)